#!/usr/bin/perl -w
#
#############################################################################
#
# File: install.pl
#
# Purpose: Installer for fwknop
#
# Credits:  (see the CREDITS file)
#
# Copyright (C) 1999-2002 Michael Rash (mbr@cipherdyne.org)
#
# License (GNU Public License):
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#    USA
#
#############################################################################
#
# $Id: install.pl,v 1.37 2005/08/06 16:32:35 mbr Exp $
#

use Cwd;
use File::Copy;
use File::Path;
use Getopt::Long;
use Sys::Hostname;
use strict;

#========================== config ===========================
my $libdir     = '/usr/lib/fwknop';
my $confdir    = '/etc/fwknop';
my $vardir     = '/var/log/fwknop';
my $varlibdir  = '/var/lib/fwknop';
my $piddir     = '/var/run/fwknop';
my $archivedir = "$confdir/archive";
my $bindir     = '/usr/bin';
my $sbindir    = '/usr/sbin';
my $initdir    = '/etc/init.d';

my $fwknop_fifo = "$varlibdir/fwknopfifo";

### system binaries
my $chkconfigCmd = '/sbin/chkconfig';
my $rcupdateCmd  = '/sbin/rc-update';  ### Gentoo
my $makeCmd      = '/usr/bin/make';
my $perlCmd      = '/usr/bin/perl';
my $gzipCmd      = '/bin/gzip';
my $killallCmd   = '/usr/bin/killall';
my $mknodCmd     = '/bin/mknod';
my $ifconfigCmd  = '/sbin/ifconfig';
#======================== end config =========================

my $client_install = 0;
my $homedir = '';
my $syslog_conf = '';

my $print_help = 0;
my $uninstall  = 0;
my $data_method = '';
my $runlevel;

### map perl modules to versions
my %required_perl_modules = (
    'Unix::Syslog' => {
        'version' => '0.100',
        'force-lib-install' => 0,
        'mod-dir' => 'Unix-Syslog'
    },
    'Net::IPv4Addr' => {
        'version' => '0.10',
        'force-lib-install' => 0,
        'mod-dir' => 'Net-IPv4Addr'
    },
    'Net::RawIP' => {
        'version' => '0.2',
        'force-lib-install' => 0,
        'mod-dir' => 'Net-RawIP'
    },
    'Net::Ping::External' => {
        'version' => '0.11',
        'force-lib-install' => 0,
        'mod-dir' => 'Net-Ping-External'
    },
    'Net::Pcap' => {
        'version' => '0.05',
        'force-lib-install' => 0,
        'mod-dir' => 'Net-Pcap'
    },
    'Crypt::Rijndael' => {
        'version' => '0.05',
        'force-lib-install' => 0,
        'mod-dir' => 'Crypt-Rijndael'
    },
    'Crypt::CBC' => {
        'version' => '2.12',
        'force-lib-install' => 0,
        'mod-dir' => 'Crypt-CBC'
    },
    'Term::ReadKey' => {
        'version' => '2.30',
        'force-lib-install' => 0,
        'mod-dir' => 'TermReadKey'
    },
    'IPTables::Parse' => {
        'version' => '0.2',
        'force-lib-install' => 1,
        'mod-dir' => 'IPTables-Parse'
    },
    'IPTables::ChainMgr' => {
        'version' => '0.2',
        'force-lib-install' => 1,
        'mod-dir' => 'IPTables-ChainMgr'
    }
);

my %cmds = (
    'make'     => $makeCmd,
    'perl'     => $perlCmd,
    'gzip'     => $gzipCmd,
    'killall'  => $killallCmd,
    'mknod'    => $mknodCmd,
    'ifconfig' => $ifconfigCmd,
);

&usage(1) unless (GetOptions(
    'client-only'   => \$client_install,  ### force client-only installation
    'uninstall'     => \$uninstall,       ### Uninstall fwknop.
    'syslog-conf=s' => \$syslog_conf,     ### specify path to syslog config file.
    'help'          => \$print_help       ### Display help.
));
&usage(0) if $print_help;

if ($< == 0 && $> == 0) {
    ### we are installing fwknop as root (so it can either act as
    ### a server or as a program that can be executed by any user
    ### on the system).
} else {
    ### we are installing as a normal user, so see if it is ok to
    ### install within the user's home directory
    my $homedir = &user_homedir_install();
    $client_install = 1;
    $libdir = "$homedir/lib/fwknop";
    $bindir = "$homedir/bin";
}

my $distro = &get_distro();

if ($distro eq 'redhat' or $distro eq 'fedora') {
    ### add chkconfig only if we are runing on a redhat distro
    $cmds{'chkconfig'} = $chkconfigCmd;
} elsif ($distro eq 'gentoo') {
    ### add rc-update if we are running on a gentoo distro
    $cmds{'rc-update'} = $rcupdateCmd;
}

### make sure the system binaries are where we expect
### them to be.
&check_commands();

my $hostname = hostname();

my $src_dir = getcwd() or die "[*] Could not get current working directory.";

unless (-d $initdir) {
    if (-d '/etc/rc.d/init.d') {
        $initdir = '/etc/rc.d/init.d';
    } elsif (-d '/etc/rc.d') {  ### for Slackware
        $initdir = '/etc/rc.d';
    } else {
        die "[*] Cannot find the init script directory, edit ",
            "the \$initdir variable.\n";
    }
}

if (not $uninstall) {
    &install();
} else {
    &uninstall();
}
exit 0;
#======================= end main ==========================

sub install() {
    print "[+] Installing fwknop on $hostname\n";

    my $init_file = '';
    my $preserve_rv = 0;
    unless ($client_install) {
        if (&ask_to_stop_fwknop()) {
            &stop_fwknop();
        }

        unless (-d '/usr/lib') {
            mkdir '/usr/lib' or die "[*] Could not mkdir /usr/lib: $!";
        }
        unless (-d '/var/run') {
            mkdir '/var/run' or die "[*] Could not mkdir /var/run: $!";
        }
        unless (-d '/var/log') {
            mkdir '/var/log' or die "[*] Could not mkdir /var/log: $!";
        }
        unless (-d $piddir) {
            mkdir $piddir or die "[*] Could not mkdir $piddir: $!";
        }
        unless (-d $varlibdir) {
            mkdir $varlibdir or die "[*] Could not mkdir $varlibdir: $!";
        }
        unless (-d $sbindir) {
            mkdir $sbindir or die "[*] Could not mkdir $sbindir: $!";
        }
    }
    unless (-d $bindir) {
        print "[+] Creating: $bindir\n";
        mkdir $bindir or die "[*] Could not mkdir $bindir: $!";
    }

    ### config directory
    unless ($client_install) {
        if (-d $confdir) {
            print "[-] Config directory $confdir already exists.\n";
        } else {
            ### Note that root will only be able to view files in
            ### /etc/fwknop since fwknop only needs to view fwknop.conf
            ### when being run as a daemon.
            print "[+] Creating config directory: $confdir\n";
            mkdir $confdir, 0500 or die "[*] Could not mkdir $confdir: $!";
        }

        if (-d $vardir) {
            print "[-] Cache directory $vardir already exists.\n";
        } else {
            print "[+] Creating cache directory: $vardir\n";
            mkdir $vardir, 0500 or die "[*] Could not mkdir $vardir: $!";
        }

        ### archive directory for previously install config files
        if (-d $archivedir) {
            print "[-] Archive directory $archivedir already exists.\n";
        } else {
            print "[+] Creating archive directory: $archivedir\n";
            mkdir $archivedir, 0500 or die "[*] Could not mkdir $archivedir: $!";
        }
    }

    print "[+] Several perl modules needed by fwknop will be installed in\n",
        "    $libdir.  Installing them here will keep the system perl\n",
        "    library tree clean.\n";

    ### make our library directory (for perl modules)
    if ($client_install) {
        if (-d "$homedir/lib") {
            print "[-] Directory $homedir/lib already exists.\n";
        } else {
            print "[+] Creating directory $homedir/lib\n";
            mkdir "$homedir/lib", 0755
                or die "[*] Could not mkdir $homedir/lib: $!";
        }
    }
    if (-d $libdir) {
        print "[-] Lib directory $libdir already exists.\n";
    } else {
        print "[+] Creating directory $libdir\n";
        mkdir $libdir, 0755 or die "[*] Could not mkdir $libdir: $!";
    }

    ### install perl modules
    for my $module (keys %required_perl_modules) {
        &install_perl_module($module);
    }

    ### special case the NetPacket::<proto> modules since the NetPacket
    ### directory is just for the base class, and we need to make sure
    ### we have each of the NetPacket::IP, NetPacket::ICMP, NetPacket::UDP,
    ### and NetPacket::TCP modules.
    chdir 'NetPacket' or die "[*] Could not chdir NetPacket directory: $!";
    unless (-e 'Makefile.PL') {
        die "[*] Your NetPacket source directory appears to be incomplete!\n",
            "    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    system "$cmds{'make'} clean" if -e 'Makefile';
    system "$cmds{'perl'} Makefile.PL PREFIX=$libdir LIB=$libdir";
    system $cmds{'make'};
#        system "$cmds{'make'} test";
    system "$cmds{'make'} install";
    chdir $src_dir or die "[*] Could not chdir $src_dir: $!";

    unless ($client_install) {
        ### create the named pipe
        unless (-e $fwknop_fifo and -p $fwknop_fifo) {
            unlink $fwknop_fifo if -e $fwknop_fifo;
            print "[+] Creating named pipe $fwknop_fifo\n";
            my $created_pipe = 1;
            unless (((system "$cmds{'mknod'} -m 600 $fwknop_fifo p")>>8) == 0) {
                $created_pipe = 0;
            }
            unless (-e $fwknop_fifo and -p $fwknop_fifo) {
                $created_pipe = 0;
            }
            unless ($created_pipe) {
                die "[*] Could not create the named pipe \"$fwknop_fifo\"!\n",
                    "[*] fwknop requires this file to exist!  Aborting install.\n";
            }
        }

        unless (-e "${vardir}/fwdata") {
            print "[+] Creating ${vardir}/fwdata file\n";
            open F, "> ${vardir}/fwdata" or die "[*] Could not open ",
                "${vardir}/fwdata: $!";
            close F;
            chmod 0600, "${vardir}/fwdata";
            &perms_ownership("${vardir}/fwdata", 0600);
        }

        ### install man pages
        &install_manpage('knopmd.8');
        &install_manpage('knopwatchd.8');
        &install_manpage('fwknop.8');

        if ($distro eq 'redhat') {
            $init_file = 'fwknop-init.redhat';
        } elsif ($distro eq 'fedora') {
            $init_file = 'fwknop-init.fedora';
        } elsif ($distro eq 'gentoo') {
            $init_file = 'fwknop-init.gentoo';
        } else {
            $init_file = 'fwknop-init.generic';
        }

        if (-e "${confdir}/fwknop.conf") {
            $preserve_rv = &query_preserve_config();
        }

        ### install all config and access files
        for my $file qw(fwknop.conf
                knopmd.conf
                knopwatchd.conf
                access.conf
                alert.conf
                pf.os) {
            if (-e "$confdir/$file") {
                &archive("$confdir/$file");
                if ($preserve_rv) {
                    if ($file eq 'access.conf') {
                        ### access.conf can have missing fields (i.e.
                        ### REQUIRE_OS_REGEX and REQUIRE_USERNAME),
                        ### and also it can have multiple sequences
                        ### defined.  Hence we just use the old one.
                        print "[+] Using original access.conf\n";
                    } elsif ($file ne 'pf.os') {
                        &preserve_config($file);
                    }
                } else {
                    print "[+] Copying $file -> $confdir\n";
                    copy $file, $confdir or
                        die "[*] Could not cp $file to $confdir";
                }
            } else {
                print "[+] Copying $file -> $confdir\n";
                copy $file, $confdir or
                    die "[*] Could not cp $file to $confdir";
            }
            if ($file eq 'knopwatchd.conf') {
                &set_hostname($file);
            }
            chmod 0600, "$confdir/$file" or
                die "[*] Could not chmod(600, $confdir/$file: $!";
            chown 0, 0, "$confdir/$file" or
                die "[*] Could not chown 0, 0, $confdir/$file: $!";
        }

        ### get data acquisition method (e.g. syslogd, sysylog-ng, ulogd
        ### or pcap)
        $data_method = &query_data_method();

        if ($data_method =~ /syslog/) {

            ### we are acquiring data via syslog
            &put_string('AUTH_MODE', 'KNOCK', "${confdir}/fwknop.conf");

            &put_string('SYSLOG_DAEMON', $data_method,
                "${confdir}/fwknop.conf");

            my $restarted_syslog = 0;
            if ($data_method eq 'syslogd') {
                if (-e $syslog_conf) {
                    &append_fifo_syslog($syslog_conf);
                    if (((system "$cmds{'killall'} -HUP syslogd " .
                            "2> /dev/null")>>8) == 0) {
                        print "[+] HUP signal sent to syslogd.\n";
                        $restarted_syslog = 1;
                    }
                }
            } elsif ($data_method eq 'syslog-ng') {
                if (-e $syslog_conf) {
                    &append_fifo_syslog_ng($syslog_conf);
                    if (((system "$cmds{'killall'} -HUP syslog-ng " .
                            "2> /dev/null")>>8) == 0) {
                        print "[+] HUP signal sent to syslog-ng.\n";
                        $restarted_syslog = 1;
                    }
                }
            }

            unless ($restarted_syslog) {
                print "[-] Could not restart any syslog daemons.\n";
            }
        } elsif ($data_method =~ /pcap/i or $data_method =~ /ulog/i) {
            ### we are using a pcap method
            if ($data_method eq 'ulogd' or $data_method eq 'file_pcap') {
                print
"[+] By default, fwknop uses the file /var/log/sniff.pcap in order to\n",
"    acquire packet data logged via a sniffer (or ulogd) to a pcap file,\n",
"    but this path may be changed by altering the PCAP_PKT_FILE keyword\n",
"    in ${confdir}/fwknop.conf.\n\n";

                if ($data_method eq 'file_pcap') {
                    &put_string('AUTH_MODE', 'FILE_PCAP',
                        "${confdir}/fwknop.conf");
                } else {
                    &put_string('AUTH_MODE', 'ULOG_PCAP',
                        "${confdir}/fwknop.conf");
                }
            } else {
                my $intf = &get_pcap_intf();
                &put_string('PCAP_INTF', $intf, "${confdir}/fwknop.conf");
            }
        } else {
            ### it is a client-only install, so don't reconfigure syslog
            ### or anything.
        }

        print "[+] Compiling knopmd and knopwatchd daemons:\n";

        ### remove any previously compiled knopmd
        unlink 'knopmd' if -e 'knopmd';

        ### remove any previously compiled knopwatchd
        unlink 'knopwatchd' if -e 'knopwatchd';

        ### compile the C fwknop daemons
        system $cmds{'make'};

        unless (-e 'knopmd' and -e 'knopwatchd') {
            die "[*] Compilation failed.";
        }

        ### install fwknop daemons/programs
        for my $daemon qw(knopmd knopwatchd knoptm) {
            print "[+] Copying $daemon -> $sbindir\n";
            copy $daemon, $sbindir or
                die "[*] Could not cp $daemon to $sbindir";
            chmod 0500, "$sbindir/$daemon" or
                die "[*] Could not chmod 500 $sbindir/$daemon";
        }
    }

    print "[+] Copying fwknop -> $bindir\n";
    copy 'fwknop', $bindir or
        die "[*] Could not cp fwknop to $bindir";

    if ($client_install) {
        open F, "< $bindir/fwknop" or die "[*] Could not open ",
            "$bindir/fwknop: $!";
        my @lines = <F>;
        close F;
        open P, "> $bindir/fwknop.tmp" or die "[*] Could not open ",
            "$bindir/fwknop.tmp: $!";
        for my $line (@lines) {
            ### change the lib dir to new homedir path
            if ($line =~ m|^\s*use\s+lib\s+\'/usr/lib/fwknop\';|) {
                print P "use lib '", $libdir, "';\n";
            } else {
                print P $line;
            }
        }
        close P;
        move "$bindir/fwknop.tmp", "$bindir/fwknop" or die "[*] Could ",
            "not move $bindir/fwknop.tmp -> $bindir/fwknop";
        chmod 0700, "$bindir/fwknop" or
            die "[*] Could not chmod 755 $bindir/fwknop";
    } else {
        chmod 0755, "$bindir/fwknop" or
            die "[*] Could not chmod 755 $bindir/fwknop";
    }

    unless ($client_install) {
        unless ($preserve_rv) {
            my $email_str = &query_email();
            if ($email_str) {
                for my $file qw(fwknop.conf knopwatchd.conf) {
                    &put_string('EMAIL_ADDRESSES', $email_str,
                        "$confdir/$file");
                }
            }
        }

        if (-d $initdir) {
            print "[+] Copying init-scripts/$init_file -> ${initdir}/fwknop\n";
            copy "init-scripts/$init_file", "${initdir}/fwknop"
                or die "[*] Could not cp fwknop init script to $initdir";
            &perms_ownership("${initdir}/fwknop", 0744);
            &enable_fwknop_at_boot($distro) unless $data_method =~ /client/i;
        }
    }

    if ($client_install) {
        print
"\n[+] fwknop has been installed at $bindir/fwknop.  Since this is a\n",
"    client-only install, the man pages could not be installed.  For more\n",
"    information about how to use fwknop, execute \"$bindir/fwknop -h\" or\n",
"    refer to:\n\n",
"        http://www.cipherdyne.org/fwknop/manpage.html\n\n";
    } else {
        if ($data_method =~ /client/i) {
            print "\n[+] fwknop has been installed!\n\n";
        } else {
            print
"\n[+] fwknop has been installed!  To start in server mode, run\n\n",
"    \"/etc/init.d/fwknop start\"\n\n",
"    Note: You will need to edit $confdir/access.conf for fwknop to\n",
"    function properly in server mode.  More information can be found in\n",
"    the fwknop manpage.\n\n";
        }
    }
    return;
}

sub uninstall() {
    ### FIXME: need to pid check to see if fwknop is still running
    ### before we remove it.
    print "[+] Uninstalling fwknop\n";

    ### stop any running fwknop daemons.
    &stop_fwknop();

    unlink "$bindir/fwknop" if -e "$bindir/fwknop";
    unlink "$sbindir/knopmd" if -e "$sbindir/knopmd";
    unlink "$sbindir/knopwatchd" if -e "$sbindir/knopwatchd";
    unlink "$initdir/fwknop" if -e "$initdir/fwknop";
    rmtree $confdir, 1, 1 if -d $confdir;
    rmtree $libdir, 1, 1 if -d $libdir;

    return;
}

### check paths to commands and attempt to correct if any are wrong.
sub check_commands() {
    my @path = qw(
        /bin
        /sbin
        /usr/bin
        /usr/sbin
        /usr/local/bin
        /usr/local/sbin
    );
    CMD: for my $cmd (keys %cmds) {
        unless (-x $cmds{$cmd}) {
            my $found = 0;
            PATH: for my $dir (@path) {
                if (-x "${dir}/${cmd}") {
                    $cmds{$cmd} = "${dir}/${cmd}";
                    $found = 1;
                    last PATH;
                }
            }
            unless ($found) {
                die "[*] Could not find $cmd anywhere!!!  ",
                    "Please edit the config section to include the path to ",
                    "$cmd.\n";
            }
        }
        unless (-x $cmds{$cmd}) {
            die "[*] $cmd is located at ",
                "$cmds{$cmd} but is not executable by uid: $<\n";
        }
    }
    return;
}

sub archive() {
    my $file = shift;
    return unless -e $file;
    my $curr_pwd = cwd();
    chdir $archivedir or die $!;
    my ($filename) = ($file =~ m|.*/(.*)|);
    my $base = "${filename}.old";
    for (my $i = 5; $i > 1; $i--) {  ### keep five copies of old config files
        my $j = $i - 1;
        unlink "${base}${i}.gz" if -e "${base}${i}.gz";
        move "${base}${j}.gz", "${base}${i}.gz" if -e "${base}${j}.gz";
    }
    print "[+] Archiving $file -> ${base}1\n";
    unlink "${base}1.gz" if -e "${base}1.gz";
    copy $file, "${base}1";   ### move $file into the archive directory
    system "$cmds{'gzip'} ${base}1";
    chdir $curr_pwd or die $!;
    return;
}

sub preserve_config() {
    my $file = shift;
    open C, "< $file" or die "[*] Could not open $file: $!";
    my @new_lines = <C>;
    close C;

    open CO, "< ${confdir}/$file" or die "[*] Could not open ",
        "${confdir}/$file: $!";
    my @orig_lines = <CO>;
    close CO;

    print "[+] Preserving existing config: ${confdir}/$file\n";
    ### write to a tmp file and then move so any running fwknop daemon will
    ### re-import a full config file if a HUP signal is received during
    ### the install.
    open CONF, "> ${confdir}/${file}.new" or die "[*] Could not open ",
        "${confdir}/${file}.new: $!";
    for my $new_line (@new_lines) {
        if ($new_line =~ /^\s*#/) {
            print CONF $new_line;  ### take comments from new file.
        } elsif ($new_line =~ /^\s*(\S+)/) {
            my $var = $1;
            my $found = 0;
            for my $orig_line (@orig_lines) {
                if ($orig_line =~ /^\s*$var\s/) {
                    print CONF $orig_line;
                    $found = 1;
                    last;
                }
            }
            unless ($found) {
                print CONF $new_line;
            }
        } else {
            print CONF $new_line;
        }
    }
    close CONF;
    move "${confdir}/${file}.new", "${confdir}/$file" or die "[*] ",
        "Could not move ${confdir}/${file}.new -> ",
        "${confdir}/$file: $!";
    return;
}

sub append_fifo_syslog() {
    print "[+] Modifying /etc/syslog.conf to write kern.info messages to\n";
    print "    $fwknop_fifo\n";
    unless (-e '/etc/syslog.conf.orig') {
        copy '/etc/syslog.conf', '/etc/syslog.conf.orig';
    }
    &archive('/etc/syslog.conf');
    open RS, '< /etc/syslog.conf' or
        die "[*]  Unable to open /etc/syslog.conf: $!\n";
    my @slines = <RS>;
    close RS;
    open SYSLOG, '> /etc/syslog.conf' or
        die "[*]  Unable to open /etc/syslog.conf: $!\n";
    for my $line (@slines) {
        unless ($line =~ /fwknopfifo/) {
            print SYSLOG $line;
        }
    }
    ### reinstate kernel logging to our named pipe
    print SYSLOG '### Send kern.info messages to fwknopfifo for ',
        "analysis by knopmd\n",
        "kern.info\t\t|$fwknop_fifo\n";
    close SYSLOG;
    return;
}

sub append_fifo_syslog_ng() {
    print "[+] Modifying /etc/syslog-ng/syslog-ng.conf to write kern.info ",
        "messages to\n";
    print "    $fwknop_fifo\n";
    unless (-e '/etc/syslog-ng/syslog-ng.conf.orig') {
        copy '/etc/syslog-ng/syslog-ng.conf',
            '/etc/syslog-ng/syslog-ng.conf.orig';
    }
    &archive('/etc/syslog-ng/syslog-ng.conf');
    open RS, '< /etc/syslog-ng/syslog-ng.conf' or
        die "[*]  Unable to open /etc/syslog-ng/syslog-ng.conf: $!\n";
    my @slines = <RS>;
    close RS;

    my $found_fifo = 0;
    for my $line (@slines) {
        $found_fifo = 1 if ($line =~ /fwknopfifo/);
    }

    unless ($found_fifo) {
        open SYSLOGNG, '>> /etc/syslog-ng/syslog-ng.conf' or
            die "[*] Unable to open /etc/syslog-ng/syslog-ng.conf: $!\n";
        print SYSLOGNG "\n",
            "destination fwknoppipe { pipe(\"/var/lib/fwknop/fwknopfifo\"); };\n",
            "filter f_kerninfo { facility(kern) and level(info); };\n",
            "log { source(src); filter(f_kerninfo); destination(fwknoppipe); };\n";
        close SYSLOGNG;
    }
    return;
}

sub get_distro() {
    return 'gentoo' if -e '/etc/gentoo-release';
    if (-e '/etc/issue') {
        ### Red Hat Linux release 6.2 (Zoot)
        open ISSUE, '< /etc/issue' or
            die "[*] Could not open /etc/issue: $!";
        my @lines = <ISSUE>;
        close ISSUE;
        for my $line (@lines) {
            chomp $line;
            return 'redhat' if $line =~ /red\s*hat/i;
            return 'fedora' if $line =~ /fedora/i;
        }
    }
    return 'NA';
}

sub install_manpage() {
    my $manpage = shift;

    ### default location to put man pages, but check with
    ### /etc/man.config
    my $mpath = '/usr/share/man/man8';
    if (-e '/etc/man.config') {
        ### prefer to install $manpage in /usr/local/man/man8 if
        ### this directory is configured in /etc/man.config
        open M, '< /etc/man.config' or
            die "[*] Could not open /etc/man.config: $!";
        my @lines = <M>;
        close M;
        ### prefer the path "/usr/share/man"
        my $found = 0;
        for my $line (@lines) {
            chomp $line;
            if ($line =~ m|^MANPATH\s+/usr/share/man|) {
                $found = 1;
                last;
            }
        }
        ### try to find "/usr/local/man" if we didn't find /usr/share/man
        unless ($found) {
            for my $line (@lines) {
                chomp $line;
                if ($line =~ m|^MANPATH\s+/usr/local/man|) {
                    $mpath = '/usr/local/man/man8';
                    $found = 1;
                    last;
                }
            }
        }
        ### if we still have not found one of the above man paths,
        ### just select the first one out of /etc/man.config
        unless ($found) {
            for my $line (@lines) {
                chomp $line;
                if ($line =~ m|^MANPATH\s+(\S+)|) {
                    $mpath = $1;
                    last;
                }
            }
        }
    }
    mkdir $mpath, 0755 unless -d $mpath;
    my $mfile = "${mpath}/${manpage}";
    print "[+] Installing $manpage man page at $mfile\n";
    copy $manpage, $mfile or die "[*] Could not copy $manpage to ",
        "$mfile: $!";
    &perms_ownership($mfile, 0644);
    print "[+] Compressing manpage $mfile\n";
    ### remove the old one so gzip doesn't prompt us
    unlink "${mfile}.gz" if -e "${mfile}.gz";
    system "$cmds{'gzip'} $mfile";
    return;
}

sub user_homedir_install() {
    my $username = '';
    my $uid = $< || die "[*] Could not get userid.";
    open P, '< /etc/passwd' or die "[*] Could not open /etc/passwd";
    while (<P>) {
        my $line = $_;
        chomp $line;
        ### mbr:x:995:101::/home/mbr:/bin/bash
        if ($line =~ /^\S+:\S+:$uid:.*:(\S+):\S+/) {
            $homedir = $1;
        }
    }
    close P;
    die "[*] Could not extract home directory from /etc/passwd."
        unless $homedir;
    print
"[+] It looks like you are installing fwknop as a non-root user.  This will\n",
"    result in fwknop being installed in your local home directory.\n",
"    Specifically, fwknop will be installed at $homedir/bin/fwknop, and a few\n",
"    perl modules needed by fwknop may be installed in $homedir/lib/fwknop/.\n",
"    If you want fwknop to act as a server (i.e. monitor knock sequences or\n",
"    authentication packets and modify firewall access or execute commands\n",
"    accordingly), you will need to install as root.\n\n";

    my $ans = '';
    while ($ans ne 'y' && $ans ne 'n') {
        print "[+] Continue with install ([y]/n)?  ";
        $ans = <STDIN>;
        if ($ans eq "\n") {  ### allow the default of "y" to take over
            $ans = 'y';
        }
        chomp $ans;
    }
    if ($ans eq 'y') {
        return $homedir;
    } else {
        die "[*] Aborting install.";
    }
}

sub enable_fwknop_at_boot() {
    my $distro = shift;
    my $ans = '';
    while ($ans ne 'y' && $ans ne 'n') {
        print "[+] Enable fwknop at boot time ([y]/n)?  ";
        $ans = <STDIN>;
        if ($ans eq "\n") {  ### allow the default of "y" to take over
            $ans = 'y';
        }
        chomp $ans;
    }
    if ($ans eq 'y') {
        if ($distro eq 'redhat' or $distro eq 'fedora') {
            system "$cmds{'chkconfig'} --add fwknop";
        } elsif ($distro eq 'gentoo') {
            system "$cmds{'rc-update'} add fwknop default";
        } else {  ### it is a non-redhat distro, try to
                  ### get the runlevel from /etc/inittab
            if ($runlevel) {
                ### the link already exists, so don't re-create it
                unless (-e "/etc/rc.d/rc${runlevel}.d/S99fwknop") {
                    symlink '/etc/rc.d/init.d/fwknop',
                        "/etc/rc.d/rc${runlevel}.d/S99fwknop";
                }
            } elsif (-e '/etc/inittab') {
                open I, '< /etc/inittab' or die "[*] Could not open ",
                    "/etc/inittab: $!";
                my @ilines = <I>;
                close I;
                for my $line (@ilines) {
                    chomp $line;
                    if ($line =~ /^id\:(\d)\:initdefault/) {
                        $runlevel = $1;
                        last;
                    }
                }
                unless ($runlevel) {
                    print "[-]  Could not determine the runlevel.  Set ",
                        "the runlevel\nmanually in the config section of ",
                        "install.pl\n";
                    return;
                }
                ### the link already exists, so don't re-create it
                unless (-e "/etc/rc.d/rc${runlevel}.d/S99fwknop") {
                    symlink '/etc/rc.d/init.d/fwknop',
                        "/etc/rc.d/rc${runlevel}.d/S99fwknop";
                }
            } else {
                print "[-] /etc/inittab does not exist!  Set the ",
                    "runlevel\nmanually in the config section of ",
                    "install.pl.\n";
                return;
            }
        }
    }
    return;
}

sub install_perl_module() {
    my $mod_name = shift;

    die '[*] Missing version key in required_perl_modules hash.'
        unless defined $required_perl_modules{$mod_name}{'version'};
    die '[*] Missing force-lib-install key in required_perl_modules hash.'
        unless defined $required_perl_modules{$mod_name}{'force-lib-install'};
    die '[*] Missing mod-dir key in required_perl_modules hash.'
        unless defined $required_perl_modules{$mod_name}{'mod-dir'};

    my $version = $required_perl_modules{$mod_name}{'version'};

    my $install_module = 0;

    if ($required_perl_modules{$mod_name}{'force-lib-install'}) {
        ### install regardless of whether the module may already be
        ### installed (this module may be a CPAN module that has been
        ### modified specifically for this project, or is a dedicated
        ### module for this project).
        $install_module = 1;
    } else {
        if (has_perl_module($mod_name)) {
            print "[+] Module $mod_name is already installed in the ",
                "system perl tree, skipping.\n";
        } else {
            ### install the module in the /usr/lib/fwknop directory because
            ### it is not already installed.
            $install_module = 1;
        }
    }

    if ($install_module) {
        unless (-d $libdir) {
            print "[+] Creating $libdir\n";
            mkdir $libdir, 0755 or die "[*] Could not mkdir $libdir: $!";
        }
        print "[+] Installing $mod_name $version perl module in $libdir/\n";
        my $mod_dir = $required_perl_modules{$mod_name}{'mod-dir'};
        chdir $mod_dir or die "[*] Could not chdir to ",
            "$mod_dir: $!";
        unless (-e 'Makefile.PL') {
            die "[*] Your $mod_name source directory appears to be incomplete!\n",
                "    Download the latest sources from ",
                "http://www.cipherdyne.org\n";
        }
        system "$cmds{'make'} clean" if -e 'Makefile';
        system "$cmds{'perl'} Makefile.PL PREFIX=$libdir LIB=$libdir";
        system $cmds{'make'};
#        system "$cmds{'make'} test";
        system "$cmds{'make'} install";
        chdir $src_dir or die "[*] Could not chdir $src_dir: $!";

        print "\n\n";
    }
    return;
}

sub set_hostname() {
    my $file = shift;
    if (-e $file) {
        open P, "< $file" or die "[*] Could not open $file: $!";
        my @lines = <P>;
        close P;
        ### replace the "HOSTNAME           CHANGE_ME" line
        open PH, "> $file" or die "[*] Could not open $file: $!";
        for my $line (@lines) {
            chomp $line;
            if ($line =~ /^\s*HOSTNAME(\s+)_?CHANGE.?ME_?/) {
                print PH "HOSTNAME${1}$hostname;\n";
            } else {
                print PH "$line\n";
            }
        }
        close PH;
    } else {
        die "[*] Your source directory appears to be incomplete!  $file ",
            "is missing.\n    Download the latest sources from ",
            "http://www.cipherdyne.org\n";
    }
    return;
}

sub query_preserve_config() {
    print "\n",
"[+] It appears that there is an existing fwknop installation on the system.\n",
"    If the existing fwknop is an older version, some configuration defaults\n",
"    may be altered in the new installation.  Note that fwknop in server mode\n",
"    now defaults to pcap mode (see the AUTH_MODE keyword in\n",
"    ${confdir}/fwknop.conf).\n\n";
    my $ans = '';
    while ($ans ne 'y' && $ans ne 'n') {
        print "    Would you like to merge the config from the existing\n",
            "    fwknop installation ([y]/n)?  ";
        $ans = <STDIN>;
        $ans = 'y' if $ans eq "\n";
        chomp $ans;
    }
    if ($ans eq 'y') {
        return 1;
    }
    return 0;
}

sub query_email() {
    my $email_str = '';
    open F, "< $confdir/fwknop.conf" or die "[*] Could not open ",
        "$confdir/fwknop.conf: $!";
    my @clines = <F>;
    close F;
    my $email_addresses;
    for my $line (@clines) {
        chomp $line;
        if ($line =~ /^\s*EMAIL_ADDRESSES\s+(.+);/) {
            $email_addresses = $1;
            last;
        }
    }
    unless ($email_addresses) {
        return '';
    }
    print "[+] fwknop access alerts will be sent to:\n\n",
        "       $email_addresses\n\n";
    my $ans = '';
    while ($ans ne 'y' and $ans ne 'n') {
        print "[+] Would you like access alerts sent to a different ",
            "address ([y]/n)?  ";
        $ans = <STDIN>;
        if ($ans eq "\n") {  ### allow the default of "y" to take over
                             ### when just "Enter" is pressed.
            $ans = 'y';
        }
        chomp $ans;
    }
    print "\n";
    if ($ans eq 'y') {
        print "\n",
"[+] To which email address(es) would you like fwknop alerts to be sent?\n",
"    You can enter as many email addresses as you like; each on its own line.\n\n",
"    End with a \".\" on a line by itself.\n\n";
        my $ans = '';
        while ($ans !~ /^\s*\.\s*$/) {
            print "    Email Address: ";
            $ans = <STDIN>;
            chomp $ans;
            if ($ans =~ m|^\s*(\S+\@\S+)$|) {
                $email_str .= "$1, ";
            } elsif ($ans !~ /^\s*\.\s*$/) {
                print "[-] Invalid email address \"$ans\"\n";
            }
        }
        $email_str =~ s/\,\s*$//;
    }
    return $email_str;
}

sub ask_to_stop_fwknop() {
    if (-e "$piddir/fwknop.pid") {
        open P, "< $piddir/fwknop.pid" or die "[*] Could not open ",
            "$piddir/fwknop.pid: $!";
        my $pid = <P>;
        close P;
        chomp $pid;
        if (kill 0, $pid) {
            print "[+] An existing fwknop process is running.\n";
            my $ans = '';
            while ($ans ne 'y' && $ans ne 'n') {
                print "    Can I stop the existing fwknop process ([y]/n)?  ";
                $ans = <STDIN>;
                if ($ans eq "\n") {  ### allow the default of "y" to take over
                    $ans = 'y';
                }
                chomp $ans;
            }
            if ($ans eq 'y') {
                return 1;
            } else {
                die "[*] Aborting install.";
            }
        }
    }
    return 0;
}

sub stop_fwknop() {
    if (-e "$piddir/fwknop.pid") {
        open P, "< $piddir/fwknop.pid" or die "[*] Could not open ",
            "$piddir/fwknop.pid: $!";
        my $pid = <P>;
        close P;
        chomp $pid;
        if (kill 0, $pid) {
            print "[+] Stopping running fwknop daemon.\n";
            system "$initdir/fwknop stop";
        }
    }
    return;
}

sub get_pcap_intf() {
    print
"\n[+] It appears that the following network interfaces are attached to the\n",
"    system:\n";
    open IFC, "$cmds{'ifconfig'} -a |" or die "[*] Could not execute ",
        "$cmds{'ifconfig'} -a: $!";
    my @ifconfig_out = <IFC>;
    close IFC;
    my %interfaces = ();
    for my $line (@ifconfig_out) {
        if ($line =~ /^\s*(\w+)\s+Link/) {
            $interfaces{$1} = '';
            print "        $1\n";
        }
    }
    my $ans = '';
    while (not defined $interfaces{$ans}) {
        print
"    Which network interface would you like fwknop to sniff packets from?  ";
        $ans = <STDIN>;
        chomp $ans;
    }
    return $ans;
}

sub query_data_method() {
    print
"[+] fwknop can act as a server (i.e. monitoring authentication packets\n",
"    and sequences, and taking the appropriate action on the local system\n",
"    to alter the firewall policy or execute commands), or as a client (i.e.\n",
"    by manufacturing authentication packets and sequences).\n\n";
    my $ans = '';

    print
"    In which mode will fwknop be executed on the local system?  (Note that\n",
"    fwknop can still be used as a client even if you select \"server\" here).\n";
    while ($ans ne 'client' and $ans ne 'server') {
        print "    (client/[server]): ";
        $ans = <STDIN>;
        if ($ans eq "\n") {
            $ans = 'server';
        }
        $ans =~ s/\s*//g;
    }
    return 'client' if $ans eq 'client';

    print
"\n[+] In server mode, fwknop can acquire packet data through syslogd or\n",
"    syslog-ng (via Netfilter logs), through a pcap file that is generated\n",
"    by a sniffer (or through the Netfilter ulogd pcap writer), or by\n",
"    sniffing packets directly off the wire via the Net::Pcap perl module.\n",
"    If you intend to use Netfilter log messages (only makes sense for the\n",
"    legacy port knocking mode), then fwknop will need to reconfigure your\n",
"    syslog daemon to write kern.info messages to the $fwknop_fifo\n",
"    named pipe.  Note that the three pcap modes provide more functionality\n",
"    and provide better security; see the fwknop man page for more\n",
"    information.\n\n";

    $ans = '';

        print
"    Which of the following data acquistion methods would you like to use?\n";
    while ($ans ne 'syslogd' and $ans ne 'syslog-ng' and $ans ne 'ulogd'
            and $ans ne 'pcap' and $ans ne 'file_pcap') {
        print "    (syslogd, syslog-ng, file_pcap, ulogd, pcap): ";
        $ans = <STDIN>;
        $ans =~ s/\s*//g;

        if ($ans eq 'syslogd' or $ans eq 'syslog-ng') {
            if ($ans eq 'syslogd') {
                ### allow command line --syslog-conf arg to take over
                $syslog_conf = '/etc/syslog.conf' unless $syslog_conf;
            } elsif ($ans eq 'syslog-ng') {
                ### allow command line --syslog-conf arg to take over
                $syslog_conf = '/etc/syslog-ng/syslog-ng.conf' unless $syslog_conf;
            }
            if ($syslog_conf and not -e $syslog_conf) {
                die
"[-] The config file $syslog_conf does not exist. Re-run install.pl\n",
"    with the --syslog-conf argument to specify the path to the syslog\n",
"    daemon config file.\n";
            }
        }
        print "[-] Invalid acquistion method \"$ans\"\n"
        unless ($ans and
            ($ans eq 'syslogd' or $ans eq 'syslog-ng'
            or $ans eq 'file_pcap' or $ans eq 'ulogd' or $ans eq 'pcap'));
    }
    return $ans;
}

sub put_string() {
    my ($var, $value, $file) = @_;
    open RF, "< $file" or die "[*] Could not open $file: $!";
    my @lines = <RF>;
    close RF;
    open F, "> $file" or die "[*] Could not open $file: $!";
    for my $line (@lines) {
        if ($line =~ /^\s*$var\s+.*;/) {
            printf F "%-28s%s;\n", $var, $value;
        } else {
            print F $line;
        }
    }
    close F;
    return;
}

sub perms_ownership() {  ### only gets called in full-installation mode
    my ($file, $perm_value) = @_;
    chmod $perm_value, $file or die "[*] Could not ",
        "chmod($perm_value, $file): $!";
    ### root (maybe should take the group assignment out?)
    chown 0, 0, $file or die "[*] Could not chown 0,0,$file: $!";
    return;
}

sub has_perl_module() {
    my $module = shift;

    # 5.8.0 has a bug with require Foo::Bar alone in an eval, so an
    # extra statement is a workaround.
    my $file = "$module.pm";
    $file =~ s{::}{/}g;
    eval { require $file };

    return $@ ? 0 : 1;
}

sub usage() {
    my $exit_status = shift;
    print "usage\n";
    exit $exit_status;
}
